import cmath
import copy
from scipy.optimize import minimize
from numpy.core.umath import exp
from scipy.misc.common import logsumexp
from scipy.special.basic import digamma
from data import *
np.random.seed(1503)

class kern_util:

    @staticmethod
    def make_kern_matrix(X,hyp,kern="se"):
        if(kern == "se"):
            K = kern_util.se_kern(X,hyp)
        elif(kern=="pe"):
            K = kern_util.se_kern(X,hyp)+kern_util.pe_kern(X,hyp)
        else:
            K = kern_util.exp_kern(X,hyp)
        return K

    @staticmethod
    def compute_kern_grad(X, hyp, kern="se"):
        if (kern == "se"):
            g_K=kern_util.se_kern_grad(X,hyp)
        elif (kern == "pe"):
            g_K = kern_util.se_kern(X, hyp) + kern_util.pe_kern(X, hyp)
        else:
            g_K = kern_util.exp_kern(X, hyp)
        return g_K

    @staticmethod
    def se_kern(X,hyp):
        K=pow(np.expand_dims(X,-1)-np.expand_dims(X,-2),2);
        if(X.ndim==2):
            K=np.sum(K,axis=0);

        K = (hyp[0] ** 2) * np.exp((-K / (hyp[1]** 2)))
        return K

    @staticmethod
    def pe_kern(X,hyp):
        K = abs(np.expand_dims(X, -1) - np.expand_dims(X, -2));
        if (X.ndim == 2):
            K = np.sum(K, axis=0);
        K = np.pow(np.sin(cmath.pi * K / (hyp[4] ** 2)),2)

        K = (hyp[2]**2) * np.exp((-K / (hyp[3] ** 2)))


        return K

    @staticmethod
    def exp_kern(X,hyp):
        K = abs(np.expand_dims(X, -1) - np.expand_dims(X, -2));
        if (X.ndim == 2):
            K = np.sum(K, axis=0);

        K = (hyp[0] ** 2) * np.exp((-K / (hyp[1] ** 2)))

        return K

    @staticmethod
    def se_kern_grad(X, hyp):
        K = pow(np.expand_dims(X, -1) - np.expand_dims(X, -2), 2);
        if (X.ndim == 2):
            K = np.sum(K, axis=0);

        g_K1 = 2*hyp[0] * np.exp((-K / (hyp[1] ** 2)))
        g_K2 = 2*(K / (hyp[1] ** 3)) * (hyp[0]**2) * np.exp((-K / (hyp[1] ** 2)))

        g_K = np.array([g_K1,g_K2])

        return g_K

    @staticmethod
    def pe_kern_grad(X, hyp):
        K = pow(np.expand_dims(X, -1) - np.expand_dims(X, -2), 2);
        if (X.ndim == 2):
            K = np.sum(K, axis=0);

        g_K1 = 2 * hyp[0] * np.exp((-K / (hyp[1] ** 2)))
        g_K2 = 2 * (K / (hyp[1] ** 3)) * (hyp[0] ** 2) * np.exp((-K / (hyp[1] ** 2)))

        K = abs(np.expand_dims(X, -1) - np.expand_dims(X, -2));
        if (X.ndim == 2):
            K = np.sum(K, axis=0);
        K = np.pow(np.sin(cmath.pi * K / (hyp[4] ** 2)), 2)

        K = (hyp[2] ** 2) * np.exp((-K / (hyp[3] ** 2)))

        g_K3 = 2 * hyp[2] * np.exp((-K / (hyp[3] ** 2)))
        g_K4 = 2 * (K / (hyp[3] ** 3)) * (hyp[2] ** 2) * np.exp((-K / (hyp[3] ** 2)))
        #fixed in real exp
        g_K5 = 0
        g_K = np.array([g_K1, g_K2,g_K3,g_K4],g_K5)

        return g_K

    @staticmethod
    def exp_kern_grad(X, hyp):
        K = np.abs(np.expand_dims(X, -1) - np.expand_dims(X, -2));
        if (X.ndim == 2):
            K = np.sum(K, axis=0);

        g_K1 = 2 * hyp[0] * np.exp((-K / (hyp[1] ** 2)))
        g_K2 = 2 * (K / (hyp[1] ** 3)) * (hyp[0] ** 2) * np.exp((-K / (hyp[1] ** 2)))

        g_K = np.array([g_K1, g_K2])

        return g_K

class gpModel:
    # hyper parameters & truncation level & others
    _eps=0.01
    _kernelParam = np.array([5,40,0,0,0])
    _sigma = 0.4
    _alpha = 1
    _truncLv1 = 15
    _numiter=40
    _kern_type = 'se'

    def __init__(self, dt=data()):
        self._numOfData = dt.Y.shape[0]
        self._numOfTimePoints = dt.Y.shape[1]
        self._tm = np.zeros(self._numOfTimePoints)
        self._kernelCov = np.zeros((self._truncLv1,self._numOfTimePoints, self._numOfTimePoints))
        self._invCov = np.zeros((self._truncLv1,self._numOfTimePoints, self._numOfTimePoints))

        self.groupKernelParam = np.zeros((self._truncLv1, 5))
        self.groupParam = np.zeros((self._numOfData, self._truncLv1))
        self.FuncTheta = np.zeros((self._truncLv1, self._numOfTimePoints))
        self.FuncVar = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))

        self.initialize_param(dt)

    def initialize_param(self, dt=data()):

        self.groupParam = np.random.dirichlet(np.ones(len(self.groupParam[0])), len(self.groupParam))

        # in simul_exp1
        self._tm = np.array(
            [0.0759666916908419, 0.123318934835166, 0.183907788282417, 0.239916153553658, 0.289952525664903,
             0.350952380892271, 0.401808033751942, 0.513249539867053, 0.622055131485066, 0.853031117721894])
        #other
        #self._tm = range(self._numOfTimePoints)+1

        K = kern_util.make_kern_matrix(self._tm, self._kernelParam,self._kern_type)
        invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))

        for i in range(self._truncLv1):
            self.groupKernelParam[i, :] = self._kernelParam
            self._kernelCov[i] = K
            self._invCov[i] = invK

    def estimate(self, dt=data()):


        a = np.zeros(self._truncLv1)
        b = np.zeros(self._truncLv1)
        temp=np.zeros((self._numOfData,self._truncLv1))

        for n in range(self._numiter):

            for k in range(self._truncLv1):
                self._kernelCov[k] = kern_util.se_kern(self._tm, self.groupKernelParam[k])
                self._invCov[k] = np.linalg.inv(self._kernelCov[k] + self._eps * np.eye(self._numOfTimePoints))

                a[k] = np.sum(self.groupParam[:, k]) + 1
                b[k] = np.sum(self.groupParam[:, k + 1:]) + self._alpha
                self.FuncVar[k] = np.linalg.inv(self._invCov[k] + (np.sum(self.groupParam[:, k]) / self._sigma) * np.eye(self._numOfTimePoints))
                self.FuncTheta[k] = np.dot(self.FuncVar[k],np.sum((self.groupParam[:, k].reshape((self._numOfData, 1))/self._sigma) * dt.Y, axis=0))
                self.groupParam[:,k]=digamma(a[k])-digamma(a[k]+b[k])+np.sum(digamma(b)[:k])-np.sum(digamma(a+b)[:k])-(self._numOfTimePoints / 2) * np.log(2 * cmath.pi * self._sigma)


            for i in range(self._numOfData):
                temp[i]=np.sum((dt.Y[i] - self.FuncTheta)*(dt.Y[i] - self.FuncTheta),axis=1) +self.FuncVar.trace(axis1=1,axis2=2)
                self.groupParam[i]-=temp[i]/ (self._sigma)
                sumGroupParam = logsumexp(self.groupParam[i])
                self.groupParam[i] = self.groupParam[i] - sumGroupParam
                self.groupParam[i] = exp(self.groupParam[i])


            self._sigma=np.sum(self.groupParam*temp)/(self._numOfTimePoints*np.sum(self.groupParam))
            self.estimate_kernel_param()


    def estimate_kernel_param(self):
        hypnum =2
        if (self._kern_type=='pe'):
            hypnum = 4
        for k in range(self._truncLv1):

            def grad(x):
                grad = np.zeros(hypnum)
                gK = kern_util.compute_kern_grad(self._tm, x, kern=self._kern_type)
                for i in range(hypnum):
                    K = kern_util.make_kern_matrix(self._tm, x, kern=self._kern_type);
                    try:
                        invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))
                    except np.linalg.LinAlgError as err:
                        invK = np.linalg.inv(K + np.eye(self._numOfTimePoints))
                    invgK = - invK.dot(gK[i]).dot(invK)
                    grad[i] = np.trace(invK.dot(gK[i])) + np.trace(invgK.dot(self.FuncVar[k])) + self.FuncTheta[k].T.dot(invgK).dot(self.FuncTheta[k])
                return grad
            def objfunc(x):
                K = kern_util.make_kern_matrix(self._tm, x, kern=self._kern_type);
                (sign, ldetK) = np.linalg.slogdet(K)
                try:
                    invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))
                except np.linalg.LinAlgError as err:
                    invK = np.linalg.inv(K + np.eye(self._numOfTimePoints))
                obj = ldetK + np.trace(np.dot(invK, self.FuncVar[k])) + self.FuncTheta[k].T.dot(invK).dot(
                    self.FuncTheta[k])
                return -obj


            #print('kern %d update'%(k))
            bound = []
            for i in range(hypnum):
                bound.append((self._eps,None))
            res = minimize(objfunc, self.groupKernelParam[k,0:hypnum], method='L-BFGS-B', jac=grad,bounds =bound,options = {'gtol': 1e-6, 'disp': False})
            self.groupKernelParam[k, 0:hypnum]=res.x

class gpModelWithSpatialCorr:

    # hyper parameters & truncation level & others
    _eps = 0.01

    _kernelParam = np.array([5, 40, 0, 0, 0])
    _spatialParam = np.array([0.1,0.01])
    _sigma = 0.4
    _alpha = 1
    _truncLv1 = 15
    _numiter = 40
    _kern_type = 'se'

    def __init__(self, dt=data()):
        self._numOfData = dt.Y.shape[0]
        self._numOfTimePoints = dt.Y.shape[1]
        self._tm = np.zeros(self._numOfTimePoints)
        self._kernelCov = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))
        self._invCov = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))

        self.groupKernelParam = np.zeros((self._truncLv1, 5))
        self.groupParam = np.zeros((self._numOfData, self._truncLv1))
        self.FuncTheta = np.zeros((self._truncLv1, self._numOfTimePoints))
        self.FuncVar = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))

        self.SpatialMu=np.zeros((self._numOfTimePoints,self._numOfData))
        self.SpatialCovMat = np.zeros((self._numOfData, self._numOfData))
        self.SpatialParam = np.zeros(2)

        self.initialize_param(dt)

    def initialize_param(self, dt=data()):

        self.groupParam = np.random.dirichlet(np.ones(len(self.groupParam[0])), len(self.groupParam))

        # in simul_exp1
        self._tm = np.array(
            [0.0759666916908419, 0.123318934835166, 0.183907788282417, 0.239916153553658, 0.289952525664903,
             0.350952380892271, 0.401808033751942, 0.513249539867053, 0.622055131485066, 0.853031117721894])
        # other
        # self._tm = range(self._numOfTimePoints)+1

        K = kern_util.make_kern_matrix(self._tm, self._kernelParam, self._kern_type)
        invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))

        for i in range(self._truncLv1):
            self.groupKernelParam[i, :] = self._kernelParam
            self._kernelCov[i] = K
            self._invCov[i] = invK
        self.SpatialParam = self._spatialParam
        self.SpatialCovMat=kern_util.make_kern_matrix(dt.S.T, self.SpatialParam, 'se')

    def estimate(self, dt=data()):

        Model1 = gpModel(dt)
        Model1._numiter=5
        Model1.estimate(dt)

        self.groupParam = copy.deepcopy(Model1.groupParam)
        self.FuncTheta = copy.deepcopy(Model1.FuncTheta)
        self.groupKernelParam = copy.deepcopy(Model1.groupKernelParam)

        a = np.zeros(self._truncLv1)
        b = np.zeros(self._truncLv1)
        temp = np.zeros((self._numOfData, self._truncLv1))

        for n in range(self._numiter):

            InvSCM = np.linalg.inv(self.SpatialCovMat)
            SpatialVar = np.linalg.inv(InvSCM + (1 / self._sigma) * np.eye(self._numOfData))

            for m in range(self._numOfTimePoints):
                self.SpatialMu[m, :] = (1 / self._sigma) * np.dot(SpatialVar, dt.Y[:, m] - np.sum(self.groupParam * self.FuncTheta[:, m], axis=1))

            for k in range(self._truncLv1):

                a[k] = np.sum(self.groupParam[:, k]) + 1
                b[k] = np.sum(self.groupParam[:, k + 1:]) + self._alpha
                self.FuncVar[k] = np.linalg.inv(self._invCov[k] + (np.sum(self.groupParam[:, k]) / self._sigma) * np.eye(self._numOfTimePoints))
                self.FuncTheta[k] = np.dot(self.FuncVar[k], np.sum((self.groupParam[:, k].reshape((self._numOfData, 1)) / self._sigma) * (dt.Y - self.SpatialMu.T),axis=0))
                self.groupParam[:, k] = digamma(a[k]) - digamma(a[k] + b[k]) + np.sum(digamma(b)[:k]) - np.sum(digamma(a + b)[:k]) - (self._numOfTimePoints / 2) * np.log(2 * cmath.pi * self._sigma)

            for i in range(self._numOfData):
                # use log_sum_exp trick
                sumDigamma = 0
                for j in range(self._truncLv1):
                    if j > 0:
                        sumDigamma += digamma(b[j - 1]) - digamma(a[j - 1] + b[j - 1])

                    self.groupParam[i][j] = digamma(a[j]) - digamma(a[j] + b[j]) - self._numOfTimePoints / 2 * np.log(
                        2 * cmath.pi * self._sigma) - (np.dot(
                        dt.Y[i] - self.FuncTheta[j] - (0.5 / self._numOfTimePoints) * self.SpatialMu[:, i],
                        dt.Y[i] - self.FuncTheta[j] - (0.5 / self._numOfTimePoints) * self.SpatialMu[:, i]) + SpatialVar[i][
                                                           i] * self._numOfTimePoints + self.FuncVar[j].trace()) / (
                                                          2 * self._sigma)

                sumGroupParam = logsumexp(self.groupParam[i])
                self.groupParam[i] = self.groupParam[i] - sumGroupParam
                self.groupParam[i] = exp(self.groupParam[i])

            for i in range(self._numOfData):
                temp[i] = (
                    np.sum((dt.Y[i] - self.FuncTheta - self.SpatialMu[:, i]) * (dt.Y[i] - self.FuncTheta - self.SpatialMu[:, i]),
                           axis=1) + self.FuncVar.trace(axis1=1, axis2=2) + self._numOfTimePoints * SpatialVar[i, i])

            sigma_spatial = 0.0
            Cor = np.linalg.inv(self.SpatialCovMat/self.SpatialParam[0])
            for m in range(self._numOfTimePoints):
                sigma_spatial += (np.trace(np.dot(Cor, SpatialVar)) + np.dot(np.dot(self.SpatialMu[m, :], Cor),
                                                                                   self.SpatialMu[m, :])) / (
                                       self._numOfTimePoints * self._numOfData)
            self._sigma = np.sum(self.groupParam * temp) / (self._numOfTimePoints * np.sum(self.groupParam))
            if (sigma_spatial > 0):
                self.SpatialParam[0] = np.sqrt(sigma_spatial)
            self.SpatialCovMat = kern_util.make_kern_matrix(dt.S.T, self.SpatialParam, 'se')
            self.estimate_kernel_param()
            self.estimate_spatial_param(dt)

    def estimate_kernel_param(self):
        hypnum = 2
        if (self._kern_type == 'pe'):
            hypnum = 4
        for k in range(self._truncLv1):

            def grad(x):
                grad = np.zeros(hypnum)
                gK = kern_util.compute_kern_grad(self._tm, x, kern=self._kern_type)
                for i in range(hypnum):
                    K = kern_util.make_kern_matrix(self._tm, x, kern=self._kern_type);
                    try:
                        invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))
                    except np.linalg.LinAlgError as err:
                        invK = np.linalg.inv(K + np.eye(self._numOfTimePoints))
                    invgK = - invK.dot(gK[i]).dot(invK)
                    grad[i] = np.trace(invK.dot(gK[i])) + np.trace(invgK.dot(self.FuncVar[k])) + self.FuncTheta[
                        k].T.dot(invgK).dot(self.FuncTheta[k])
                return grad

            def objfunc(x):
                K = kern_util.make_kern_matrix(self._tm, x, kern=self._kern_type);
                (sign, ldetK) = np.linalg.slogdet(K)
                try:
                    invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))
                except np.linalg.LinAlgError as err:
                    invK = np.linalg.inv(K + np.eye(self._numOfTimePoints))
                obj = ldetK + np.trace(np.dot(invK, self.FuncVar[k])) + self.FuncTheta[k].T.dot(invK).dot(
                    self.FuncTheta[k])
                return -obj

            # print('kern %d update'%(k))
            bound = []
            for i in range(hypnum):
                bound.append((self._eps, None))
            res = minimize(objfunc, self.groupKernelParam[k, 0:hypnum], method='L-BFGS-B', jac=grad, bounds=bound,
                           options={'gtol': 1e-6, 'disp': False})
            self.groupKernelParam[k, 0:hypnum] = res.x

    def estimate_spatial_param(self,dt):
        def grad(x):
            hypset = np.array([self.SpatialParam[0],x])
            grad = 0;
            gK = kern_util.compute_kern_grad(dt.S.T, hypset, kern='se')
            K = kern_util.make_kern_matrix(dt.S.T, hypset, kern='se');

            invK = np.linalg.inv(K + 0.01*np.eye(self._numOfData))
            invgK = - invK.dot(gK[1]).dot(invK)
            grad = np.trace(invK.dot(gK[1])) + np.trace(invgK.dot(self.SpatialCovMat))
            temp = 0
            for j in range(self._numOfTimePoints):
                temp = temp + self.SpatialMu[j,:].T.dot(invgK).dot(self.SpatialMu[j,:])
            grad = grad + (1/self._numOfTimePoints) * temp
            return grad

        def objfunc(x):
            hypset = np.array([self.SpatialParam[0], x])
            K = kern_util.make_kern_matrix(dt.S.T, hypset, kern='se');
            (sign, ldetK) = np.linalg.slogdet(K)

            invK = np.linalg.inv(K + 0.01*np.eye(self._numOfData))
            obj = ldetK + np.trace(np.dot(invK, self.SpatialCovMat))
            temp = 0
            for j in range(self._numOfTimePoints):
                temp = temp + self.SpatialMu[j, :].T.dot(invK).dot(self.SpatialMu[j, :])
            obj = obj + (1/self._numOfTimePoints) * temp
            return -obj

        res = minimize(objfunc, self.SpatialParam[1], method='BFGS', jac=grad)
        self.SpatialParam[1] = res.x

class ProposedModel:
    # hyper parameters & truncation level & others
    _eps = 0.01
    _kernelParam = np.array([5, 40, 0, 0,0])
    _spatialParam = np.array([0.1, 0.01])
    _sigma = 0.4
    _alpha = 1
    _truncLv1 = 15
    _truncLv2 = 7
    _numiter = 40
    _kern_type = 'se'
    _beta = 1

    def __init__(self, dt=data()):

        self._numOfData = dt.Y.shape[0]
        self._numOfTimePoints = dt.Y.shape[1]
        self._tm = np.zeros(self._numOfTimePoints)
        self._kernelCov = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))
        self._invCov = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))

        self.groupKernelParam = np.zeros((self._truncLv1, 5))
        self.groupParam = np.zeros((self._numOfData, self._truncLv1))
        self.FuncTheta = np.zeros((self._truncLv1, self._numOfTimePoints))
        self.FuncVar = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))

        self.SpatialMu = np.zeros((self._numOfTimePoints, self._numOfData))
        self.SpatialCovMat = np.zeros((self._numOfData, self._numOfData))
        self.SpatialParam = np.zeros(2)

        self.locGroupParam = np.zeros((self._numOfData, self._truncLv1, self._truncLv2))
        self.locTheta = np.zeros((self._truncLv1, self._truncLv2, 2))
        self.loctau = np.zeros((self._truncLv1, self._truncLv2))
        self.locVar = np.zeros((self._truncLv1, self._truncLv2, 2, 2))
        self.locpsi = np.zeros((self._truncLv1, self._truncLv2))

        self._mu_zero = np.zeros(2)
        self._tau = 1
        self._Lambda = np.eye(dt.S.shape[1])
        self._psi = 3

        self.initialize_param(dt)

    def initialize_param(self, dt=data()):

        self.groupParam = np.random.dirichlet(np.ones(len(self.groupParam[0])), len(self.groupParam))

        for i in range(self._numOfData):
            self.locGroupParam[i] = np.random.dirichlet(np.ones(len(self.locGroupParam[0][0])),len(self.locGroupParam[0]))

        # in simul_exp1
        self._tm = np.array(
            [0.0759666916908419, 0.123318934835166, 0.183907788282417, 0.239916153553658, 0.289952525664903,
             0.350952380892271, 0.401808033751942, 0.513249539867053, 0.622055131485066, 0.853031117721894])
        # other
        # self._tm = range(self._numOfTimePoints)+1

        K = kern_util.make_kern_matrix(self._tm, self._kernelParam, self._kern_type)
        invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))

        for i in range(self._truncLv1):
            self.groupKernelParam[i, :] = self._kernelParam
            self._kernelCov[i] = K
            self._invCov[i] = invK
        self.SpatialParam = self._spatialParam
        self.SpatialCovMat = kern_util.make_kern_matrix(dt.S.T, self.SpatialParam, 'se')

    def estimate(self, dt=data()):

        a = np.zeros(self._truncLv1)
        b = np.zeros(self._truncLv1)
        c = np.zeros((self._truncLv1, self._truncLv2))
        d = np.zeros((self._truncLv1, self._truncLv2))
        temp_spatial = np.zeros((self._numOfData, self._truncLv1))

        Model1 = gpModel(dt)
        Model1._numiter = 5
        Model1.estimate(dt)

        self.groupParam = copy.deepcopy(Model1.groupParam)
        self.FuncTheta = copy.deepcopy(Model1.FuncTheta)
        self.groupKernelParam = copy.deepcopy(Model1.groupKernelParam)

        for n in range(self._numiter):

            InvSCM = np.linalg.inv(self.SpatialCovMat)
            SpatialVar = np.linalg.inv(InvSCM + (1 / self._sigma) * np.eye(self._numOfData))

            for m in range(self._numOfTimePoints):
                self.SpatialMu[m, :] = (1 / self._sigma) * np.dot(SpatialVar, dt.Y[:, m] - np.sum(
                    self.groupParam * self.FuncTheta[:, m], axis=1))

            for k in range(self._truncLv1):
                a[k] = np.sum(self.groupParam[:, k]) + 1
                b[k] = np.sum(self.groupParam[:, k + 1:]) + self._alpha
                self.FuncVar[k] = np.linalg.inv(
                    self._invCov[k] + (np.sum(self.groupParam[:, k]) / self._sigma) * np.eye(self._numOfTimePoints))
                self.FuncTheta[k] = np.dot(self.FuncVar[k], np.sum(
                    (self.groupParam[:, k].reshape((self._numOfData, 1)) / self._sigma) * (dt.Y - self.SpatialMu.T),
                    axis=0))
                #self.groupParam[:, k] = digamma(a[k]) - digamma(a[k] + b[k]) + np.sum(digamma(b)[:k]) - np.sum(
                #    digamma(a + b)[:k]) - (self._numOfTimePoints / 2) * np.log(2 * cmath.pi * self._sigma)

            for i in range(self._truncLv1):
                for j in range(self._truncLv2):

                    temp = self.groupParam[:, i]*self.locGroupParam[:, i, j]
                    sPsi = np.sum(temp)+self._eps
                    tempS= np.sum(temp * np.array(dt.S).T,axis=1)

                    self.loctau[i,j] = self._tau + sPsi
                    self.locTheta[i,j] = (self._tau*self._mu_zero+tempS) / self.loctau[i,j]
                    self.locpsi[i,j] = self._psi + sPsi + 1
                    tempVar = np.sum(temp*(np.expand_dims(dt.S-tempS/sPsi,-1) * np.expand_dims(dt.S-tempS/sPsi,-2)).T,axis=2)
                    self.locVar[i, j] = np.linalg.inv(np.linalg.inv(self._Lambda) + ((self._tau*sPsi)/self._tau+sPsi) * np.dot((tempS/sPsi-self._mu_zero) , (tempS/sPsi-self._mu_zero).T) + tempVar)

                    c[i][j] = sPsi + 1
                    d[i][j] = np.sum(np.dot(self.groupParam[:, i], np.sum(self.locGroupParam[:, i, j + 1:], axis=1))) + self._beta

            #groupPram update
            for i in range(self._numOfData):

                sumDigamma = 0
                for j in range(self._truncLv1):
                    if j > 0:
                        sumDigamma += digamma(b[j - 1]) - digamma(a[j - 1] + b[j - 1])
                    self.groupParam[i][j] = digamma(a[j]) - digamma(a[j] + b[j]) - self._numOfTimePoints / 2 * np.log(2 * cmath.pi * self._sigma)- (np.dot(dt.Y[i] - self.FuncTheta[j]-(0.5/self._numOfTimePoints)*self.SpatialMu[:,i], dt.Y[i] - self.FuncTheta[j]-(0.5/self._numOfTimePoints)*self.SpatialMu[:,i])+SpatialVar[i][i]*self._numOfTimePoints +self.FuncVar[j].trace()) / (2 * self._sigma)
                    self.groupParam[i][j] += sumDigamma
                    B=self.groupParam[i][j]
                    sumDigamma2 = 0

                    for k in range(self._truncLv2):
                        if k > 0:
                            sumDigamma2 += digamma(d[j][k - 1]) - digamma(c[j][k - 1] + d[j][k - 1])

                        (s, logDet)=np.linalg.slogdet(self.locVar[j,k])
                        E_q_Lambda = logDet + digamma(self.locpsi[j,k] / 2) + digamma((self.locpsi[j,k] -1) / 2) + 2 * np.log(2)
                        temp_L =  (0.5) *E_q_Lambda - (1/self.loctau[j,k]) - (self.locpsi[j,k]/2) * np.dot((dt.S[i] - self.locTheta[j,k]).T,self.locVar[j,k]).dot((dt.S[i] - self.locTheta[j,k])) -np.log(2*cmath.pi)
                        self.groupParam[i][j] += self.locGroupParam[i][j][k] * ((digamma(c[j][k]) - digamma(c[j][k] + d[j][k])) + temp_L + sumDigamma2)




                sumGroupParam = logsumexp(self.groupParam[i])
                self.groupParam[i] = self.groupParam[i] - sumGroupParam
                self.groupParam[i] = exp(self.groupParam[i])

            #locgroupPram update
            for i in range(self._numOfData):
                for j in range(self._truncLv1):
                    sumDigamma2 = 0
                    for k in range(self._truncLv2):
                        if k > 0:
                            sumDigamma2 += digamma(d[j][k - 1]) - digamma(c[j][k - 1] + d[j][k - 1])

                        self.locGroupParam[i][j][k] = digamma(c[j][k]) - digamma(c[j][k] + d[j][k])


                        (s, logDet) = np.linalg.slogdet(self.locVar[j, k])
                        E_q_Lambda = logDet + digamma(self.locpsi[j,k] / 2) + digamma((self.locpsi[j,k] - 1) / 2)
                        temp_L = (-0.5) * E_q_Lambda - (1 / self.loctau[j, k]) - (self.locpsi[j, k] / 2) * np.dot(
                            (dt.S[i] - self.locTheta[j, k]).T, self.locVar[j, k]).dot((dt.S[i] - self.locTheta[j, k])) -np.log(2*cmath.pi)

                        self.locGroupParam[i][j][k] += temp_L
                        self.locGroupParam[i][j][k] += sumDigamma2
                    sumlocGroupParam = logsumexp(self.locGroupParam[i][j])
                    self.locGroupParam[i][j] = self.locGroupParam[i][j] - sumlocGroupParam
                    self.locGroupParam[i][j] = exp(self.locGroupParam[i][j])

            for i in range(self._numOfData):
                temp_spatial[i] = (
                    np.sum((dt.Y[i] - self.FuncTheta - self.SpatialMu[:, i]) * (
                    dt.Y[i] - self.FuncTheta - self.SpatialMu[:, i]),
                           axis=1) + self.FuncVar.trace(axis1=1, axis2=2) + self._numOfTimePoints * SpatialVar[i, i])

            sigma_spatial = 0.0
            Cor = np.linalg.inv(self.SpatialCovMat / self.SpatialParam[0])
            for m in range(self._numOfTimePoints):
                sigma_spatial += (np.trace(np.dot(Cor, SpatialVar)) + np.dot(np.dot(self.SpatialMu[m, :], Cor),
                                                                             self.SpatialMu[m, :])) / (
                                     self._numOfTimePoints * self._numOfData)
            self._sigma = np.sum(self.groupParam * temp_spatial) / (self._numOfTimePoints * np.sum(self.groupParam))
            if (sigma_spatial > 0):
                self.SpatialParam[0] = np.sqrt(sigma_spatial)
            self.SpatialCovMat = kern_util.make_kern_matrix(dt.S.T, self.SpatialParam, 'se')
            self.estimate_kernel_param()
            self.estimate_spatial_param(dt)

    def estimate_kernel_param(self):
        hypnum = 2
        if (self._kern_type == 'pe'):
            hypnum = 4
        for k in range(self._truncLv1):

            def grad(x):
                grad = np.zeros(hypnum)
                gK = kern_util.compute_kern_grad(self._tm, x, kern=self._kern_type)
                for i in range(hypnum):
                    K = kern_util.make_kern_matrix(self._tm, x, kern=self._kern_type);
                    try:
                        invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))
                    except np.linalg.LinAlgError as err:
                        invK = np.linalg.inv(K + np.eye(self._numOfTimePoints))
                    invgK = - invK.dot(gK[i]).dot(invK)
                    grad[i] = np.trace(invK.dot(gK[i])) + np.trace(invgK.dot(self.FuncVar[k])) + self.FuncTheta[
                        k].T.dot(invgK).dot(self.FuncTheta[k])
                return grad

            def objfunc(x):
                K = kern_util.make_kern_matrix(self._tm, x, kern=self._kern_type);
                (sign, ldetK) = np.linalg.slogdet(K)
                try:
                    invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))
                except np.linalg.LinAlgError as err:
                    invK = np.linalg.inv(K + np.eye(self._numOfTimePoints))
                obj = ldetK + np.trace(np.dot(invK, self.FuncVar[k])) + self.FuncTheta[k].T.dot(invK).dot(
                    self.FuncTheta[k])
                return -obj

            # print('kern %d update'%(k))
            bound = []
            for i in range(hypnum):
                bound.append((self._eps, None))
            res = minimize(objfunc, self.groupKernelParam[k, 0:hypnum], method='L-BFGS-B', jac=grad, bounds=bound,
                           options={'gtol': 1e-6, 'disp': False})
            self.groupKernelParam[k, 0:hypnum] = res.x

    def estimate_spatial_param(self, dt):
        def grad(x):
            hypset = np.array([self.SpatialParam[0], x])
            grad = 0;
            gK = kern_util.compute_kern_grad(dt.S.T, hypset, kern='se')
            K = kern_util.make_kern_matrix(dt.S.T, hypset, kern='se');

            invK = np.linalg.inv(K + np.eye(self._numOfData))
            invgK = - invK.dot(gK[1]).dot(invK)
            grad = np.trace(invK.dot(gK[1])) + np.trace(invgK.dot(self.SpatialCovMat))
            temp = 0
            for j in range(self._numOfTimePoints):
                temp = temp + self.SpatialMu[j, :].T.dot(invgK).dot(self.SpatialMu[j, :])
            grad = grad + (1 / self._numOfTimePoints) * temp
            return grad

        def objfunc(x):
            hypset = np.array([self.SpatialParam[0], x])
            K = kern_util.make_kern_matrix(dt.S.T, hypset, kern='se');
            (sign, ldetK) = np.linalg.slogdet(K)

            invK = np.linalg.inv(K + np.eye(self._numOfData))
            obj = ldetK + np.trace(np.dot(invK, self.SpatialCovMat))
            temp = 0
            for j in range(self._numOfTimePoints):
                temp = temp + self.SpatialMu[j, :].T.dot(invK).dot(self.SpatialMu[j, :])
            obj = obj + (1 / self._numOfTimePoints) * temp
            return -obj

        res = minimize(objfunc, self.SpatialParam[1], method='BFGS', jac=grad)
        self.SpatialParam[1] = res.x

class gpModelLoc:
            # hyper parameters & truncation level & others
            _eps = 0.01
            _kernelParam = np.array([5, 40, 0, 0, 0])

            _sigma = 0.4
            _alpha = 1
            _truncLv1 = 15
            _truncLv2 = 7
            _numiter = 40
            _kern_type = 'se'
            _beta = 1

            def __init__(self, dt=data()):

                self._numOfData = dt.Y.shape[0]
                self._numOfTimePoints = dt.Y.shape[1]
                self._tm = np.zeros(self._numOfTimePoints)
                self._kernelCov = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))
                self._invCov = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))

                self.groupKernelParam = np.zeros((self._truncLv1, 5))
                self.groupParam = np.zeros((self._numOfData, self._truncLv1))
                self.FuncTheta = np.zeros((self._truncLv1, self._numOfTimePoints))
                self.FuncVar = np.zeros((self._truncLv1, self._numOfTimePoints, self._numOfTimePoints))

                self.locGroupParam = np.zeros((self._numOfData, self._truncLv1, self._truncLv2))
                self.locTheta = np.zeros((self._truncLv1, self._truncLv2, 2))
                self.loctau = np.zeros((self._truncLv1, self._truncLv2))
                self.locVar = np.zeros((self._truncLv1, self._truncLv2, 2, 2))
                self.locpsi = np.zeros((self._truncLv1, self._truncLv2))

                self._mu_zero = np.zeros(2)
                self._tau = 1
                self._Lambda = np.eye(dt.S.shape[1])
                self._psi = 3

                self.initialize_param(dt)

            def initialize_param(self, dt=data()):


                self.groupParam = np.random.dirichlet(np.ones(len(self.groupParam[0])), len(self.groupParam))

                for i in range(self._numOfData):
                    self.locGroupParam[i] = np.random.dirichlet(np.ones(len(self.locGroupParam[0][0])),
                                                                len(self.locGroupParam[0]))

                # in simul_exp1
                self._tm = np.array(
                    [0.0759666916908419, 0.123318934835166, 0.183907788282417, 0.239916153553658, 0.289952525664903,
                     0.350952380892271, 0.401808033751942, 0.513249539867053, 0.622055131485066, 0.853031117721894])
                # other
                # self._tm = range(self._numOfTimePoints)+1

                K = kern_util.make_kern_matrix(self._tm, self._kernelParam, self._kern_type)
                invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))

                for i in range(self._truncLv1):
                    self.groupKernelParam[i, :] = self._kernelParam
                    self._kernelCov[i] = K
                    self._invCov[i] = invK

            def estimate(self, dt=data()):

                a = np.zeros(self._truncLv1)
                b = np.zeros(self._truncLv1)
                c = np.zeros((self._truncLv1, self._truncLv2))
                d = np.zeros((self._truncLv1, self._truncLv2))
                temp2 = np.zeros((self._numOfData, self._truncLv1))

                Model1 = gpModel(dt)
                Model1._numiter = 5
                Model1.estimate(dt)

                self.groupParam = copy.deepcopy(Model1.groupParam)
                self.FuncTheta = copy.deepcopy(Model1.FuncTheta)
                self.groupKernelParam = copy.deepcopy(Model1.groupKernelParam)

                for n in range(self._numiter):

                    for k in range(self._truncLv1):
                        self._kernelCov[k] = kern_util.se_kern(self._tm, self.groupKernelParam[k])
                        self._invCov[k] = np.linalg.inv(self._kernelCov[k] + self._eps * np.eye(self._numOfTimePoints))

                        a[k] = np.sum(self.groupParam[:, k]) + 1
                        b[k] = np.sum(self.groupParam[:, k + 1:]) + self._alpha
                        self.FuncVar[k] = np.linalg.inv(
                            self._invCov[k] + (np.sum(self.groupParam[:, k]) / self._sigma) * np.eye(
                                self._numOfTimePoints))
                        self.FuncTheta[k] = np.dot(self.FuncVar[k], np.sum(
                            (self.groupParam[:, k].reshape((self._numOfData, 1)) / self._sigma) * dt.Y, axis=0))

                    for i in range(self._truncLv1):
                        for j in range(self._truncLv2):
                            temp = self.groupParam[:, i] * self.locGroupParam[:, i, j]
                            sPsi = np.sum(temp) + self._eps
                            tempS = np.sum(temp * np.array(dt.S).T, axis=1)

                            self.loctau[i, j] = self._tau + sPsi
                            self.locTheta[i, j] = (self._tau * self._mu_zero + tempS) / self.loctau[i, j]
                            self.locpsi[i, j] = self._psi + sPsi + 1
                            tempVar = np.sum(temp * (
                            np.expand_dims(dt.S - tempS / sPsi, -1) * np.expand_dims(dt.S - tempS / sPsi, -2)).T,
                                             axis=2)
                            self.locVar[i, j] = np.linalg.inv(
                                np.linalg.inv(self._Lambda) + ((self._tau * sPsi) / self._tau + sPsi) * np.dot(
                                    (tempS / sPsi - self._mu_zero), (tempS / sPsi - self._mu_zero).T) + tempVar)

                            c[i][j] = sPsi + 1
                            d[i][j] = np.sum(np.dot(self.groupParam[:, i],
                                                    np.sum(self.locGroupParam[:, i, j + 1:], axis=1))) + self._beta
                    # groupPram update
                    for i in range(self._numOfData):

                        sumDigamma = 0
                        for j in range(self._truncLv1):
                            if j > 0:
                                sumDigamma += digamma(b[j - 1]) - digamma(a[j - 1] + b[j - 1])
                            self.groupParam[i][j] = digamma(a[j]) - digamma(
                                a[j] + b[j]) - self._numOfTimePoints / 2 * np.log(2 * cmath.pi * self._sigma) - (np.dot(
                                dt.Y[i] - self.FuncTheta[j],
                                dt.Y[i] - self.FuncTheta[j]) +self.FuncVar[j].trace()) / (2 * self._sigma)
                            self.groupParam[i][j] += sumDigamma

                            sumDigamma2 = 0

                            for k in range(self._truncLv2):
                                if k > 0:
                                    sumDigamma2 += digamma(d[j][k - 1]) - digamma(c[j][k - 1] + d[j][k - 1])

                                (s, logDet) = np.linalg.slogdet(self.locVar[j, k])
                                E_q_Lambda = logDet + digamma(self.locpsi[j, k] / 2) + digamma(
                                    (self.locpsi[j, k] - 1) / 2) + 2 * np.log(2)
                                temp_L = (0.5) * E_q_Lambda - (1 / self.loctau[j, k]) - (
                                                                                        self.locpsi[j, k] / 2) * np.dot(
                                    (dt.S[i] - self.locTheta[j, k]).T, self.locVar[j, k]).dot(
                                    (dt.S[i] - self.locTheta[j, k])) - np.log(2 * cmath.pi)


                                self.groupParam[i][j] += self.locGroupParam[i][j][k] * (
                                (digamma(c[j][k]) - digamma(c[j][k] + d[j][k])) + temp_L + sumDigamma2)


                        sumGroupParam = logsumexp(self.groupParam[i])
                        self.groupParam[i] = self.groupParam[i] - sumGroupParam
                        self.groupParam[i] = exp(self.groupParam[i])
                    # locgroupPram update
                    for i in range(self._numOfData):
                        for j in range(self._truncLv1):
                            sumDigamma2 = 0
                            for k in range(self._truncLv2):
                                if k > 0:
                                    sumDigamma2 += digamma(d[j][k - 1]) - digamma(c[j][k - 1] + d[j][k - 1])

                                self.locGroupParam[i][j][k] = digamma(c[j][k]) - digamma(c[j][k] + d[j][k])

                                (s, logDet) = np.linalg.slogdet(self.locVar[j, k])
                                E_q_Lambda = logDet + digamma(self.locpsi[j, k] / 2) + digamma(
                                    (self.locpsi[j, k] - 1) / 2)
                                temp_L = (-0.5) * E_q_Lambda - (1 / self.loctau[j, k]) - (self.locpsi[
                                                                                              j, k] / 2) * np.dot(
                                    (dt.S[i] - self.locTheta[j, k]).T, self.locVar[j, k]).dot(
                                    (dt.S[i] - self.locTheta[j, k])) - np.log(2 * cmath.pi)

                                self.locGroupParam[i][j][k] += temp_L
                                self.locGroupParam[i][j][k] += sumDigamma2
                            sumlocGroupParam = logsumexp(self.locGroupParam[i][j])
                            self.locGroupParam[i][j] = self.locGroupParam[i][j] - sumlocGroupParam
                            self.locGroupParam[i][j] = exp(self.locGroupParam[i][j])

                    for i in range(self._numOfData):
                        temp2[i] = np.sum((dt.Y[i] - self.FuncTheta) * (dt.Y[i] - self.FuncTheta), axis=1) + self.FuncVar.trace(axis1=1, axis2=2)


                    self._sigma = np.sum(self.groupParam * temp2) / (self._numOfTimePoints * np.sum(self.groupParam))
                    self.estimate_kernel_param()
                    # self.estimate_spatial_param(dt)

            def estimate_kernel_param(self):
                hypnum = 2
                if (self._kern_type == 'pe'):
                    hypnum = 4
                for k in range(self._truncLv1):

                    def grad(x):
                        grad = np.zeros(hypnum)
                        gK = kern_util.compute_kern_grad(self._tm, x, kern=self._kern_type)
                        for i in range(hypnum):
                            K = kern_util.make_kern_matrix(self._tm, x, kern=self._kern_type);
                            try:
                                invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))
                            except np.linalg.LinAlgError as err:
                                invK = np.linalg.inv(K + np.eye(self._numOfTimePoints))
                            invgK = - invK.dot(gK[i]).dot(invK)
                            grad[i] = np.trace(invK.dot(gK[i])) + np.trace(invgK.dot(self.FuncVar[k])) + self.FuncTheta[
                                k].T.dot(invgK).dot(self.FuncTheta[k])
                        return grad

                    def objfunc(x):
                        K = kern_util.make_kern_matrix(self._tm, x, kern=self._kern_type);
                        (sign, ldetK) = np.linalg.slogdet(K)
                        try:
                            invK = np.linalg.inv(K + self._eps * np.eye(self._numOfTimePoints))
                        except np.linalg.LinAlgError as err:
                            invK = np.linalg.inv(K + np.eye(self._numOfTimePoints))
                        obj = ldetK + np.trace(np.dot(invK, self.FuncVar[k])) + self.FuncTheta[k].T.dot(invK).dot(
                            self.FuncTheta[k])
                        return -obj

                    # print('kern %d update'%(k))
                    bound = []
                    for i in range(hypnum):
                        bound.append((self._eps, None))
                    res = minimize(objfunc, self.groupKernelParam[k, 0:hypnum], method='L-BFGS-B', jac=grad,
                                   bounds=bound,
                                   options={'gtol': 1e-6, 'disp': False})
                    self.groupKernelParam[k, 0:hypnum] = res.x




